/*
 *************************************************************************
 *                                                                       *
 * Copyright 2000 Compaq Computer Corporation                            *
 *                                                                       *
 * COMPAQ Registered in U.S. Patent and Trademark Office.                *
 *                                                                       *
 *************************************************************************
 * IMPORTANT: Carefully read the License Terms below before              *
 * proceeding.  By use of these materials you agree to these terms.      *
 * If you do not agree to these terms, you may not use this software or  *
 * the accompanying documentation.                                       *
 *************************************************************************
 * LICENSE TERMS                                                         *
 * 1. GRANT                                                              *
 * Compaq Computer Corporation ("COMPAQ") grants you the right to use,   *
 * modify, and distribute the following source code (the "Software")     *
 * on any number of computers. You may use the Software as part of       *
 * creating a software program or product intended for commercial or     *
 * non-commercial distribution in machine-readable source code, binary,  *
 * or executable formats. You may distribute the Software as             *
 * machine-readable source code provided this license is not removed     *
 * from the Software and any modifications are conspicuously indicated.  *
 * 2. COPYRIGHT                                                          *
 * The Software is owned by COMPAQ and its suppliers and is protected by *
 * copyright laws and international treaties.  Your use of the Software  *
 * and associated documentation is subject to the applicable copyright   *
 * laws and the express rights and restrictions of these terms.          *
 * 3. RESTRICTIONS                                                       *
 * You may not remove any copyright, trademark, or other proprietary     *
 * notices from the Software or the associated  documentation.           *
 * You are responsible for compliance with all applicable export or      *
 * re-export control laws and regulations if you export the Software.    *
 * This license is governed by and is to be construed under the laws     *
 * of the State of Texas.                                                *
 *                                                                       *
 * DISCLAIMER OF WARRANTY AND LIABILITY                                  *
 * Compaq shall not be liable for technical or editorial errors or       *
 * omissions contained herein. The information contained herein is       *
 * subject to change without notice.                                     *
 *                                                                       *
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND.       *
 * THE ENTIRE RISK ARISING OUT OF THE USE OF THIS SOFTWARE REMAINS WITH  *
 * RECIPIENT.  IN NO EVENT SHALL COMPAQ BE LIABLE FOR ANY DIRECT,        *
 * CONSEQUENTIAL, INCIDENTAL, SPECIAL, PUNITIVE OR OTHER DAMAGES         *
 * WHATSOEVER (INCLUDING WITHOUT LIMITATION DAMAGES FOR LOSS OF BUSINESS *
 * PROFITS, BUSINESS INTERRUPTION, OR LOSS OF BUSINESS INFORMATION),     *
 * EVEN IF COMPAQ HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES    *
 * AND WHETHER IN AN ACTION OF CONTRACT OR TORT INCLUDING NEGLIGENCE.    *
 *                                                                       *
 * If you have any questions concerning this license, please contact:    *
 * Compaq Computer Corporation, Software Business Practices, ZKO1-2/D22, *
 * 110 Spit Brook Road, Nashua, NH. 03062-2698.                          *
 *                                                                       *
 *************************************************************************
 */

/*

If you build this module normally you will just get a socketpair routine.

If you define GENERIC_SOCKETPAIRTEST, you'll also get a main that will test
the routine:

DECVMS7> cc /pref=all /define="GENERIC_SOCKETPAIRTEST" socketpair
DECVMS7> link  socketpair
DECVMS7> run socketpair
11 bytes send on socket #5
11 bytes received on socket #4
data: -->testing ...<--
DECVMS7>


  This is how the 2 threads play with each others to get the socket pair

         thread 1                              thread 2
            +
            |
       sd1=socket()
            |
         bind(..port=INADDR_ANY..)
            assume now system gives us
            port 5555 to use 
            |          
            |
        pthread_create() --------------------------+
            |                                      |
         listen()                             pthread_detach()
            |                                       |
            |                                 sd3= socket()
            |                                       |
        eventPost(up) <------ synch ------->   eventWait(Up)
            |                                       |
 +--------non-blocking select(sd1+1,..)             |
 |          (timeout= n seconds)                    |
 |          |                                       |
 |          |                                       |
 |          |                                       |
 | port=5555|<--------------------------------- connect(sd3,port=5555)
 |          |                                       |
 |     sd2= accept(sd1) <==connection established==>| (sd3)
 |          |                                       |
 |          |                                       |  
 |          |                         __________> eventPost(down)
 |      eventWait(down) <----- synch-/              |
 |          |                                   pthread_exit(0)
 |          |      
 |     Here, the pair of sockets
 |     sd2 and sd3 shown above are 
 |     our socketpair. return 0.
 |  
 +---------> timeout error. return -1.
             (this assumes second thread did
              not make it).
      

*/

/*
** We may be built with "names=as_is", so we need to uppercase all the
** pthread routine names. 
*/
#define pthread_create		PTHREAD_CREATE 
#define pthread_exit		PTHREAD_EXIT 
#define pthread_mutex_init	PTHREAD_MUTEX_INIT 
#define pthread_cond_init	PTHREAD_COND_INIT 
#define pthread_cond_destroy	PTHREAD_COND_DESTROY 
#define pthread_mutex_destroy	PTHREAD_MUTEX_DESTROY 
#define pthread_mutex_lock	PTHREAD_MUTEX_LOCK 
#define pthread_mutex_unlock	PTHREAD_MUTEX_UNLOCK 
#define pthread_cond_wait	PTHREAD_COND_WAIT 
#define pthread_cond_broadcast	PTHREAD_COND_BROADCAST 
#define pthread_detach          PTHREAD_DETACH

#include "vms_jackets.h"
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <inet.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <ssdef.h>

typedef struct {
   pthread_cond_t    cond ;
   pthread_mutex_t   mp ;
   pthread_mutex_t   mtx ;
   int               cnt ;
} eventType ;

typedef struct {
   eventType      *up ;
   eventType      *down ;
   short          port ;
   int            *sd ;
   int            fam ;
   int            type ;
} tDataType ;


static int GENERIC_SOCKETPAIR_inet ( int fam, int type, int proto, int sd[2] );
static void *GENERIC_SOCKETPAIR_inet_client ( void *p ) ;


int GENERIC_SOCKETPAIR ( int fam, int type, int proto, int sd[2] ) ;  
static void *GENERIC_SOCKETPAIR_inet_client ( void *p ) ;
static pthread_mutex_t Netscape_socketpair_lock = PTHREAD_MUTEX_INITIALIZER;
#define SOCKETPAIR_LOCK pthread_mutex_lock(&Netscape_socketpair_lock)
#define SOCKETPAIR_UNLOCK pthread_mutex_unlock(&Netscape_socketpair_lock)

static void eventInit ( eventType *e ) ;
static int  eventWait ( eventType *e ) ;
static void eventPost ( eventType *e ) ;
static void eventDestroy ( eventType *e ) ;


/*------------- start of test section --------------------------------------*/

#ifdef GENERIC_SOCKETPAIRTEST

int main ( int argc, char *argv[], char *envp[] )
{
   int   sda[2] ;

   if ( GENERIC_SOCKETPAIR ( AF_INET, SOCK_STREAM, 0, sda ) < 0 ) {
      perror ( "socketpair" ) ;
      exit ( 1 ) ;
   }

   /*
   * use socket pair
   */
   {
      char     *sb = "testing ..." ;
      char     rb[1024] ;
      int      i ;
      int      bl ;

      bl = strlen ( sb ) + 1 ;
      i = send ( sda[0], sb, bl, 0 ) ;
      printf ( "%d bytes send on socket #%d\n", i, sda[0] ) ;
      bl = sizeof ( rb ) ;
      i = recv ( sda[1], rb, bl, 0 ) ;
      printf ( "%d bytes received on socket #%d\n", i, sda[1] ) ;
      printf ( "data: -->%s<--\n", rb ) ;
   }


   close ( sda[0] ) ;
   close ( sda[1] ) ;
}
#endif /* GENERIC_SOCKETPAIRTEST */

#ifdef FULL_GENERIC_SOCKETPAIRTEST

#define NUMBER_OF_THREADS 100
#define PRINTF_STUFF 0  /* make it zero if you dont want to see printf */
#define PRINTF_DATA 1 /* to make it print the data itself being send */
#define MSG_LEN 50
static int counter;
static void* main_( void*);

int main ( int argc, char *argv[], char *envp[] )
{
   int i;
   pthread_t p;
   int stat;


   for(i=0;i < NUMBER_OF_THREADS ; i++){
      stat = pthread_create ( &p, 0, main_ , (void *)0 ) ;
      if( stat == -1){
          perror("pthread_create");
          exit(1);
      }
      else{
         if( PRINTF_STUFF ){
             fprintf(stderr, "created thread number %d\n",i);
         }
      }
      sched_yield();
   }

   for(;;)
     sched_yield();
}

void* main_( void* foo)  /*Threads function */
{
   int   sda[2] ;
   char     *sb = (char *) malloc(MSG_LEN+1);
   int i;
   char     rb[MSG_LEN+1] ;
   int      bl ;
   int status;

   status = pthread_detach( pthread_self());
   if( status != 0) {
       perror("pthread_detach");
       exit(1);
   }

   if( PRINTF_STUFF )
       fprintf(stderr, "entering main..\n");

   for(i=0;i<MSG_LEN;i++) sb[i]=' ';
   sb[MSG_LEN]='\0';

   for(;;)
   {
      if( PRINTF_STUFF ){
          counter++;
          fprintf(stderr,"==============> counter is %d\n",counter);
          fprintf(stderr,"before calling GENERIC_SOCKETPAIR()\n");
       }

       if( GENERIC_SOCKETPAIR ( AF_INET, SOCK_STREAM, 0, sda ) < 0 ){
           perror ( "GENERIC_SOCKETPAIR:main_" ) ;  
           exit(0);
       }
       else{

           if( PRINTF_STUFF )
               fprintf(stderr, 
                       "socket pair created..#%d and #%d\n",sda[0],sda[1]);

           bl = strlen ( sb ) + 1 ;

           memset(sb,(int)'N',strlen(sb));

           i = send ( sda[0], sb, bl, 0 ) ;
           bl = sizeof ( rb ) ;

           i = recv ( sda[1], rb, bl, 0 ) ;
           if( PRINTF_DATA )
               fprintf(stderr,"%s\n",rb);

           if( PRINTF_STUFF)
               fprintf(stderr, "closing socket %d  and %d \n",sda[0],sda[1]);


           close ( sda[0] ) ;
           close ( sda[1] ) ;
           sched_yield();
      }
 }/* forever */

}
#endif /* FULL_GENERIC_SOCKETPAIRTEST */

/*----------------------------- end of test section -------------------*/


/************************************************************************
 *
 * Main entry to this module.
 * 
 ************************************************************************/
int GENERIC_SOCKETPAIR ( int fam, int type, int proto, int sd[2] )
{
int status;

   switch ( fam ) {
       case AF_INET :
       case AF_UNIX :

           SOCKETPAIR_LOCK; 
                status =  GENERIC_SOCKETPAIR_inet ( AF_INET, type, proto, sd ) ;
           SOCKETPAIR_UNLOCK; 

           return status;
           break ;

         default :
           break ;
   }
}

/**********************************************************************
 * this runs in a pthread context, it will create a pthread also 
 * returns 0 on success, -1 on fail 
 **********************************************************************/
static int GENERIC_SOCKETPAIR_inet ( int fam, int type, int proto, int sd[2] )
{
   pthread_t            pt ;
   struct sockaddr_in   sin ;
   int                  sd1,
                        sd2,
                        sd3 ;
   int                  rv ;
   u_int                l ;
   short                eph_port ;
   eventType            up, down ;
   tDataType            tdata ;
   fd_set               rfds;
   int                  found;
   struct timeval       time;

   const time_t         MAX_SECONDS_TO_WAIT = 20;
   int i;
   int error_found, keep_looking;

   eventInit ( &up ) ;
   eventInit ( &down ) ; 

   if (( sd1 = socket ( fam, type, 0 )) < 0 ) {
        perror("GENERIC_SOCKETPAIR_inet: socket");
        return ( -1 ) ;
   }

   memset ((char *)&sin, 0, sizeof ( sin )) ;
   sin.sin_family = fam ;
   sin.sin_addr.s_addr = INADDR_ANY ;
   sin.sin_port = 0 ;
   l = sizeof ( sin ) ;

   if (( rv = bind ( sd1, (struct sockaddr *)&sin, l )) < 0 ) {
         perror("GENERIC_SOCKETPAIR_inet: bind");
         close( sd1 );   
         return ( -1 ) ;
   }

   if (( rv = getsockname ( sd1, (struct sockaddr *)&sin, &l )) < 0 ) {
         perror("GENERIC_SOCKETPAIR_inet: getsockname");
         close( sd1 ); 
         return ( -1 ) ;
   }

   eph_port = sin.sin_port ;

   tdata.up = &up ;
   tdata.down = &down ;
   tdata.port = eph_port ;
   tdata.sd = &sd3 ;
   tdata.fam = fam ;
   tdata.type = type ;

   if(0!=pthread_create ( &pt, 0, GENERIC_SOCKETPAIR_inet_client, (void *)&tdata)){
         perror("GENERIC_SOCKETPAIR_inet: pthread_create()");
         close(sd1);
         return -1;
   }

   listen ( sd1, 5 ) ;

   memset ((char *)&sin, 0, sizeof ( sin )) ;
   l = sizeof ( sin ) ;
   eventPost ( &up ) ; 

   FD_ZERO(&rfds);
   FD_SET(sd1, &rfds);
   time.tv_sec  = MAX_SECONDS_TO_WAIT;
   time.tv_usec = 0;
  
   error_found = FALSE;
   keep_looking= TRUE;
                                                     
   for(; keep_looking ; ){
        found = select(sd1+1, (fd_set *)&rfds, (fd_set *)0,
                     (fd_set *)0, &time);

        if(  found < 0 ){
             if(  errno != EINTR ){
                  perror("GENERIC_SOCKETPAIR_inet:select");
                  keep_looking= FALSE;
                  error_found = TRUE;
             }
        }
        else{
             if(found == 0 ){
                   fprintf(stderr,"GENERIC_SOCKETPAIR_inet: select timeout!\n");
                   vaxc$errno = SS$_TIMEOUT;
                   errno      = EVMSERR;  
                   keep_looking = FALSE;
                   error_found  = TRUE;
             }
             else{
                   keep_looking = FALSE;
             }
        }
   }
         
   if( !error_found ){
        sd2 = accept ( sd1, (struct sockaddr *)&sin, &l ) ;
        if( sd2 < 0 ){
            perror("GENERIC_SOCKETPAIR_inet:accept");
            error_found = TRUE;
        }
        else
            eventWait ( &down );  /* ignore status here.. we are leaving*/ 
   }

   close ( sd1 ) ;
   eventDestroy ( &up ) ;
   eventDestroy ( &down ) ;

   sd[0] = sd2 ;
   sd[1] = sd3 ;

   if(! error_found )
        return ( 0 ) ;
   else
        return -1;

}

/**********************************************************************
 *
 *  Thread below. This thread is created by  GENERIC_SOCKETPAIR_inet 
 *
 *********************************************************************/
static void *GENERIC_SOCKETPAIR_inet_client ( void *p )
{
   tDataType            *tdata = (tDataType *)p ;
   struct sockaddr_in   sin ;
   int status;
   static const char* local_host_ip = "127.0.0.1";

   status = pthread_detach( pthread_self());
   if( status != 0 ){
       perror("GENERIC_SOCKETPAIR_inet_client: pthread_detach"); 
   }
   else{
        status = eventWait ( tdata->up ) ;
        if( status ==0 ){
              if (( *(tdata->sd) = socket ( tdata->fam, tdata->type, 0 )) < 0 ){
                     perror("GENERIC_SOCKETPAIR_inet_client: socket"); 
              }
              else{
                   memset ((char *)&sin, 0, sizeof ( sin )) ;
                   sin.sin_family = tdata->fam ;
                   sin.sin_addr.s_addr = inet_addr ( local_host_ip ) ;
                   sin.sin_port = tdata->port ;
           
                   if( connect( *(tdata->sd), 
                                 (struct sockaddr *)&sin, 
                                 sizeof ( sin )) < 0 ){
                         perror("GENERIC_SOCKETPAIR_inet_client: connect");
                         close( *(tdata->sd) );
                   }
             }
       
             eventPost ( tdata->down ) ;
        }
   }

   pthread_exit ( 0 ) ;
   return NULL; /* keep the compiler happy */
}

/******************************************************************
 *      eventInit
 *****************************************************************/
static void eventInit ( eventType *e )
{
   if( 0 != pthread_mutex_init ( &(e->mtx), 0 )){
         perror("GENERIC_SOCKETPAIR:eventInit");
   }

   if( 0 != pthread_mutex_init ( &(e->mp), 0 )){
         perror("GENERIC_SOCKETPAIR:eventInit");
   }

   if( 0 != pthread_cond_init ( &(e->cond), 0 )){
         perror("GENERIC_SOCKETPAIR:eventInit");
   }
   e->cnt = 0 ;
}

/******************************************************************
 *
 *          eventDestroy
 *******************************************************************/
static void eventDestroy ( eventType *e )
{
   if( 0 !=pthread_cond_destroy ( &(e->cond) )){
         perror("GENERIC_SOCKETPAIR:eventDestroy");
   }

   if( 0 != pthread_mutex_destroy ( &(e->mp) )){
         perror("GENERIC_SOCKETPAIR:eventDestroy");
   }

   if( 0 != pthread_mutex_destroy ( &(e->mtx) )){
         perror("GENERIC_SOCKETPAIR:eventDestroy");
   }

}

/*****************************************************************
 *
 *            eventWait
 * return 0 on success, else -1
 ****************************************************************/
static int eventWait ( eventType *e )
{
   int doWait ;

   if( 0 != pthread_mutex_lock ( &(e->mtx) )){
         perror("GENERIC_SOCKETPAIR:eventWait");
         return -1;
   }

   doWait = ( e->cnt <= 0 ) ;

   if( 0 != pthread_mutex_unlock ( &(e->mtx) )){
         pthread_mutex_unlock ( &(e->mp) ) ;
         perror("GENERIC_SOCKETPAIR:eventWait");
         return -1;
   }

   if( doWait ) {
       if( 0 != pthread_mutex_lock ( &(e->mp) )){
            pthread_mutex_unlock ( &(e->mp) ) ;
            perror("GENERIC_SOCKETPAIR:eventWait");
            return -1;
       }

       if( 0 != pthread_cond_wait ( &(e->cond), &(e->mp) )  ) {
             pthread_mutex_unlock ( &(e->mp) ) ;
             perror("GENERIC_SOCKETPAIR:eventWait");
             return -1;
       }
       else
             pthread_mutex_unlock ( &(e->mp) ) ;
   }

   return 0;
}

/*******************************************************************
 *
 *              eventPost
 ******************************************************************/
static void eventPost ( eventType *e )
{
   if( 0 !=pthread_mutex_lock ( &(e->mtx) )){
          perror("GENERIC_SOCKETPAIR:eventPost");
          return;
   }

   e->cnt++ ;

   if( 0 != pthread_mutex_lock ( &(e->mp) )){
         pthread_mutex_unlock ( &(e->mtx) ) ;
         perror("GENERIC_SOCKETPAIR:eventPost");
         return;
   }

   if ( 0 != pthread_cond_broadcast ( &(e->cond) ) ) {
          perror("GENERIC_SOCKETPAIR:eventPost");
   }

   pthread_mutex_unlock ( &(e->mp) ) ;
   pthread_mutex_unlock ( &(e->mtx) ) ;
}
